@@ -13,6 +13,8 @@ from typing import Any, Generic, Iterator, Union
 import lmdb
 import pickle
 import json
+import pathlib
+import inspect
 
 imports = {}
 try:
@@ -101,9 +103,9 @@ class Lz4:
     def __init__(self, compress_level):
         self.compress_level = compress_level
     def compress(self, obj):
-        return zstd.compress(obj, self.compress_level)
+        return lz4.frame.compress(obj, self.compress_level)
     def decompress(self, obj):
-        return zstd.decompress(obj)
+        return lz4.frame.decompress(obj)
 
 
 #######################################################
@@ -112,60 +114,92 @@ class Lz4:
 
 class Shock(MutableMapping):
 
-    def __init__(self, file_path: str, flag: str = "r", map_size: int = 2**40, lock: bool = False, sync: bool = False, max_readers: int = 126, serializer: str = None, protocol: int = 5, compressor: str = None, compress_level: int = 1):
+    def __init__(self, file_path: str, flag: str = "r", sync: bool = False, lock: bool = True, serializer = None, protocol: int = 5, compressor = None, compress_level: int = 1, map_size: int = 2**40, **kwargs):
         """
 
         """
         ## Open lmdb
         if flag == "r":  # Open existing database for reading only (default)
-            env = lmdb.open(file_path, map_size=map_size, max_dbs=0, readonly=True, create=False, subdir=False, lock=lock, sync=False, max_readers=max_readers)
+            env = lmdb.open(file_path, map_size=map_size, max_dbs=0, readonly=True, create=False, subdir=False, lock=lock, sync=False, **kwargs)
             write = False
+            fp_exists = True
         elif flag == "w":  # Open existing database for reading and writing
-            env = lmdb.open(file_path, map_size=map_size, max_dbs=0, readonly=False, create=False, subdir=False, lock=lock, sync=sync, max_readers=max_readers)
+            env = lmdb.open(file_path, map_size=map_size, max_dbs=0, readonly=False, create=False, subdir=False, lock=lock, sync=sync, **kwargs)
             write = True
+            fp_exists = True
         elif flag == "c":  # Open database for reading and writing, creating it if it doesn't exist
-            env = lmdb.open(file_path, map_size=map_size, max_dbs=0, readonly=False, create=True, subdir=False, lock=lock, sync=sync, max_readers=max_readers)
+            fp = pathlib.Path(file_path)
+            fp_exists = fp.exists()
+
+            env = lmdb.open(file_path, map_size=map_size, max_dbs=0, readonly=False, create=True, subdir=False, lock=lock, sync=sync, **kwargs)
             write = True
         elif flag == "n":  # Always create a new, empty database, open for reading and writing
             utils.remove_db(file_path)
-            env = lmdb.open(file_path, map_size=map_size, max_dbs=0, readonly=False, create=True, subdir=False, lock=lock, sync=sync, max_readers=max_readers)
+            env = lmdb.open(file_path, map_size=map_size, max_dbs=0, readonly=False, create=True, subdir=False, lock=lock, sync=sync, **kwargs)
             write = True
+            fp_exists = False
         else:
             raise ValueError("Invalid flag")
 
         self.env = env
         self._write = write
 
-        ## Serializer
-        if serializer is None:
-            self._serializer = None
-        elif serializer == 'pickle':
-            self._serializer = Pickle(protocol)
-        elif serializer == 'json':
-            self._serializer = Json
-        elif serializer == 'orjson':
-            if imports['orjson']:
-                self._serializer = Orjson
-            else:
-                raise ValueError('orjson could not be imported.')
+        ## Load or assign encodings
+        if fp_exists:
+            with env.begin(write=False, buffers=False) as txn:
+                self._serializer = pickle.loads(txn.get(b'00~._serializer'))
+                self._compressor = pickle.loads(txn.get(b'01~._compressor'))
         else:
-            raise ValueError('serializer must be one of pickle, json, or orjson.')
+            ## Serializer
+            if serializer is None:
+                self._serializer = None
+            elif serializer == 'pickle':
+                self._serializer = Pickle(protocol)
+            elif serializer == 'json':
+                self._serializer = Json
+            elif serializer == 'orjson':
+                if imports['orjson']:
+                    self._serializer = Orjson
+                else:
+                    raise ValueError('orjson could not be imported.')
+            elif inspect.isclass(serializer):
+                class_methods = dir(serializer)
+                if ('dumps' in class_methods) and ('loads' in class_methods):
+                    self._serializer = serializer
+                else:
+                    raise ValueError('If a class is passed for a serializer, then it must have dumps and loads methods.')
+            else:
+                raise ValueError('serializer must be one of pickle, json, orjson, or a serializer class with dumps and loads methods.')
 
-        ## Compressor
-        if compressor is None:
-            self._compressor = None
-        elif compressor == 'gzip':
-            self._compressor = Gzip(compress_level)
-        elif compressor == 'zstd':
-            if imports['zstd']:
-                self._compressor = Zstd(compress_level)
+            ## Compressor
+            if compressor is None:
+                self._compressor = None
+            elif compressor == 'gzip':
+                self._compressor = Gzip(compress_level)
+            elif compressor == 'zstd':
+                if imports['zstd']:
+                    self._compressor = Zstd(compress_level)
+                else:
+                    raise ValueError('zstd could not be imported.')
+            elif compressor == 'lz4':
+                if imports['lz4']:
+                    self._compressor = Lz4(compress_level)
+                else:
+                    raise ValueError('lz4 could not be imported.')
+            elif inspect.isclass(compressor):
+                class_methods = dir(compressor)
+                if ('compress' in class_methods) and ('decompress' in class_methods):
+                    self._compressor = compressor(compress_level)
+                else:
+                    raise ValueError('If a class is passed for a compressor, then it must have compress and decompress methods as well as a compress_level parameter in the __init__.')
             else:
-                raise ValueError('zstd could not be imported.')
-        elif compressor == 'lz4':
-            if imports['lz4']:
-                self._compressor = Lz4(compress_level)
-            else:
-                raise ValueError('lz4 could not be imported.')
+                raise ValueError('compressor must be one of gzip, zstd, lz4, or a compressor class with compress and decompress methods.')
+
+            ## Save encodings if new file
+            with env.begin(write=True, buffers=False) as txn:
+                txn.put(b'00~._serializer', pickle.dumps(self._serializer, protocol))
+                txn.put(b'01~._compressor', pickle.dumps(self._compressor, protocol))
+            env.sync()
 
 
     def info(self) -> dict:
@@ -182,7 +216,6 @@ class Shock(MutableMapping):
         """
         Change the map size of the database by a value.
         """
-
         self.env.set_mapsize(value)
 
     def copy(self, file_path, compact=True):
@@ -198,7 +231,6 @@ class Shock(MutableMapping):
         """
         self.env.copy(file_path, compact=compact)
 
-
     def _pre_key(self, key: str) -> bytes:
 
         return key.encode()
@@ -257,19 +289,22 @@ class Shock(MutableMapping):
     def keys(self):
 
         with self.env.begin(write=False, buffers=False) as txn:
-            for key in txn.cursor().iternext(keys=True, values=False):
+            cursor = utils.move_cursor(txn)
+            for key in cursor.iternext(keys=True, values=False):
                 yield self._post_key(key)
 
     def items(self):
 
         with self.env.begin(write=False, buffers=False) as txn:
-            for key, value in txn.cursor().iternext(keys=True, values=True):
-                yield (self._post_key(key), self._post_value(value))
+            cursor = utils.move_cursor(txn)
+            for key, value in cursor.iternext(keys=True, values=True):
+                yield self._post_key(key), self._post_value(value)
 
     def values(self):
 
         with self.env.begin(write=False, buffers=False) as txn:
-            for value in txn.cursor().iternext(keys=False, values=True):
+            cursor = utils.move_cursor(txn)
+            for value in cursor.iternext(keys=False, values=True):
                 yield self._post_value(value)
 
     def __contains__(self, key: str) -> bool:
@@ -314,19 +349,14 @@ class Shock(MutableMapping):
             raise ValueError('File is open for read only.')
 
 
-    # The method "drop" should work for this without iterating through all objects
     def clear(self):
         if self._write:
             with self.env.begin(write=True, buffers=False) as txn:
-                for key in txn.cursor().iternext(keys=True, values=False):
+                cursor = utils.move_cursor(txn)
+                for key in cursor.iternext(keys=True, values=False):
                     txn.delete(key)
         else:
             raise ValueError('File is open for read only.')
-        # if self._write:
-        #     with self.env.begin(write=True, buffers=False) as txn:
-        #         txn.drop(db=None, delete=delete)
-        # else:
-        #     raise ValueError('File is open for read only.')
 
 
     def sync(self) -> None:
@@ -338,7 +368,7 @@ class Shock(MutableMapping):
 
     def close(self) -> None:
         """
-        Close the environment, invalidating non-synced writes.
+        Close the environment. All writes will be synced.
         """
         self.sync()
         self.env.close()
@@ -351,37 +381,48 @@ class Shock(MutableMapping):
 
 
 def open(
-    file_path: str, flag: str = "r", map_size: int = 2**40, lock: bool = False, sync: bool = False, max_readers: int = 126, serializer: str = None, protocol: int = 5, compressor: str = None, compress_level: int = 1):
+    file_path: str, flag: str = "r", sync: bool = False, lock: bool = True, serializer = None, protocol: int = 5, compressor = None, compress_level: int = 1, map_size: int = 2**40, **kwargs):
     """
-    Open a persistent dictionary for reading and writing.
+    Open a persistent dictionary for reading and writing. On creation of the file, the encodings (serializer and compressor) will be written to the file. Any reads and new writes do not need to be opened with the encoding parameters. Currently, ShockDB uses pickle to serialize the encodings to the file.
 
     Parameters
     -----------
     file_path : str or pathlib.Path
         It must be a path to a local file location.
+
     flag : str
         Flag associated with how the file is opened according to the dbm style. See below for details.
-    map_size : int
-        Maximum size database may grow to; used to size the memory mapping. If database grows larger than map_size, an exception will be raised and the user must close and reopen Environment. On 64-bit there is no penalty for making this huge (say 1TB). Must be <2GB on 32-bit.
-    lock : bool
-        If False, don’t do any locking. If concurrent access is anticipated, the caller must manage all concurrency itself. For proper operation the caller must enforce single-writer semantics, and must ensure that no readers are using old transactions while a writer is active. The simplest approach is to use an exclusive lock so that no readers may be active at all when a writer begins.
+
     sync : bool
         If False, don’t flush system buffers to disk when committing a transaction. This optimization means a system crash can corrupt the database or lose the last transactions if buffers are not yet flushed to disk.
 
-The risk is governed by how often the system flushes dirty buffers to disk and how often sync() is called. However, if the filesystem preserves write order, transactions exhibit ACI (atomicity, consistency, isolation) properties and only lose D (durability). I.e. database integrity is maintained, but a system crash may undo the final transactions.
-    max_readers : int
-        Maximum number of simultaneous read transactions. Can only be set by the first process to open an environment, as it affects the size of the lock file and shared memory area. Attempts to simultaneously start more than this many read transactions will fail.
-    serializer : str or None
+        The risk is governed by how often the system flushes dirty buffers to disk and how often sync() is called. However, if the filesystem preserves write order, transactions exhibit ACI (atomicity, consistency, isolation) properties and only lose D (durability). I.e. database integrity is maintained, but a system crash may undo the final transactions.
+
+    lock : bool
+        If False, don’t do any locking. If concurrent access is anticipated, the caller must manage all concurrency itself. For proper operation the caller must enforce single-writer semantics, and must ensure that no readers are using old transactions while a writer is active. The simplest approach is to use an exclusive lock so that no readers may be active at all when a writer begins.
+
+    serializer : str, class, or None
         The serializer to use to convert the input object to bytes. Currently, must be one of pickle, json, orjson, or None. If the objects can be serialized to json, then use orjson. It's super fast and you won't have the pickle issues.
         If None, then the input values must be bytes.
+        A class with dumps and loads methods can also be passed as a custom serializer.
+
     protocol : int
         The pickle protocol to use.
-    compressor : str or None
+
+    compressor : str, class, or None
         The compressor to use to compress the pickle object before being written. Currently, only zstd is accepted.
         The amount of compression will vary wildly depending on the input object and the serializer used. It's definitely worth doing some testing before using a compressor. Saying that...if you serialize to json, you'll likely get a lot of benefit from a fast compressor.
+        A class with compress and decompress methods can also be passed as a custom serializer. The class also needs a compress_level parameter in the __init__.
+
     compress_level : int
         The compression level for the compressor.
 
+    map_size : int
+        Maximum size database may grow to; used to size the memory mapping. If database grows larger than map_size, an exception will be raised and the user must close and reopen Environment. On 64-bit there is no penalty for making this huge (say 1TB). Must be <2GB on 32-bit.
+
+    **kwargs:
+        Any kwargs that can be passed to the lmdb.Environment class: https://lmdb.readthedocs.io/en/release/#environment-class
+
     Returns
     -------
     Shock
@@ -406,10 +447,14 @@ The risk is governed by how often the system flushes dirty buffers to disk and h
 
     """
 
-    return Shock(file_path, flag, map_size, lock, sync, max_readers, serializer, protocol, compressor, compress_level)
+    return Shock(file_path, flag, sync, lock, serializer, protocol, compressor, compress_level, map_size, **kwargs)
 
 
 
+# with open(os.path.join(tmp.name, 'test_file'), 'wb') as f:
+#           f.write(b'test')
+
+
 # for key, value in shock0.items():
 #     print(key)
 
